open Ast

let block_start="Symbol_Table_Node* parent=_CurSymbolTable;\nSymbol_Table_Node* _CurSymbolTable=new Symbol_Table_Node(parent);\n"
let block_end="delete _CurSymbolTable;\n"
(*types used to construct a environment*)
type record={name:string;start_str:string;end_str:string;var_type:vari_type;}
type env={global:record list;local:record list;func_count:int;func_code:string;is_global:bool;}

(*below functions are used for searching through environments*)
let rec search_env_list var env_list=match env_list with
	h::t->if h.name=var then h.name,h.start_str,h.end_str,h.var_type,true else search_env_list var t	
	|[]->"@","@","@",Void,false
	
let search_var var env=(*search variables in environment*)
	let a,b,c,d,e=search_env_list var env.local in
		if e=false then (search_env_list var env.global) else a,b,c,d,e
and add_var var env=(*add variables to environment*)
	if env.is_global=true then {global=var::env.global;local=env.local;func_count=env.func_count;func_code=env.func_code;is_global=env.is_global}
	else {global=env.global;local=var::env.local;func_count=env.func_count;func_code=env.func_code;is_global=env.is_global}

let concat a b= a^b
let str_of_op= function(*convert operators to string*)
	Add->"+"
	|Sub->"-"
	|Mul->"*"
	|Div->"/"
	|Mod->"%"
	|Eq->"=="
	|Neq->"!="
	|Le->">="
	|Se->"<="
	|Larger->">"
	|Smaller->"<"
	|And->"&&"
	|Or->"||"
let str_of_vartype=function(*convert types to string*)
	Int->"int"
	|Str->"string"
	|Flo->"double"
	|Bol->"bool"
	|Lin->"Line*"
	|Cir->"Circle*"
	|Poi->"Point*"
	|Void->"void"
let str_of_vartype_2=function(*convert types to string*)
	Int->"int"
	|Str->"string"
	|Flo->"double"
	|Bol->"bool"
	|Lin->"Line*"
	|Cir->"Circle*"
	|Poi->"Point*"
	|Void->"void"
let str_of_vartype_call=function(*convert types to string*)
	Int->"*((int*"
	|Str->"*((string*"
	|Flo->"*((double*"
	|Bol->"*((bool*"
	|Lin->"(*(Line**"
	|Cir->"(*(Circle**"
	|Poi->"(*(Point**"
	|Void->"((void"
let rec eval_stmt stmt env=match stmt with
	Expr(e)->(eval e env)^";\n",env
	|Return(e)-> "_return_value="^(eval e env)^";\n goto endline;\n",env
	|If(e,s1,s2,s3)->"if("^(eval e env)^"){\n"^(let x,y =eval_stmt_list (List.rev s1) env in block_start^x^block_end)^"}\n"^(let x,y =eval_stmt_list (List.rev s2) env in x)^"else{\n"^(let x,y =eval_stmt_list (List.rev s3) env in block_start^x^block_end)^"}\n",env
	|For(e1,e2,e3,s1)->"for("^(eval e1 env)^";"^(eval e2 env)^";"^(eval e3 env)^"){\n"^(let x,y =eval_stmt_list (List.rev s1) env in block_start^x^block_end)^"}\n",env
	|While(e,s)->"While("^(eval e env)^"){\n"^(let x,y =eval_stmt_list (List.rev s) env in block_start^x^block_end)^"}\n",env
	|Var_dec(v)->let x,y=(eval_var_dec v env) in x^";\n",y
	|Func_dec(f)->let x,y=(eval_fun_dec f env)in x^"\n",y
	|Break-> "break;\n",env
	|Continue->"continue;\n",env
	|Pass->"\n",env
	|Elif(e,s)->"else if("^(eval e env)^"){\n"^block_start^(let x,y =eval_stmt_list (List.rev s) env in x)^block_end^"}\n",env
and eval_stmt_list stmt_list env=match stmt_list with
	h::t->let code1,env1 =eval_stmt h env in let code2,env2 = eval_stmt_list t env1 in code1^code2, env2
	(*|[_]->eval_stmt (List.hd stmt_list) env*)
	|_->"",env
and eval_fun_dec func env=(**)
	let fun_env={global=env.global;local=[];func_count=env.func_count;func_code=env.func_code;is_global=false} in(*create a new environment for function*)
	let var_req_str=eval_var_req func.return_type
	and f_name=(ignore (print_endline func.fname));func.fname(*function name*)
	and str_dec,str_poi,str_head,new_env=eval_actual func.formals fun_env in(*str_dec:(int a, int b) str_poi:(int,int) str_head: symbol_table->add_var()*)
	let last_env={global=new_env.global;local={name=f_name;start_str="("^var_req_str^"(*";end_str=")("^str_poi^"))";var_type=Int;}::new_env.local;func_count=new_env.func_count+1;func_code=new_env.func_code;is_global=new_env.is_global}
	in let stmts,env2=eval_stmt_list  func.body last_env
	in "_CurSymbolTable->add_function((func_pointer) &function"^(string_of_int (new_env.func_count+1))^",\""^f_name^"\");\n",
	{global=if env.is_global then {name=f_name;start_str="("^var_req_str^"(*";end_str=")("^str_poi^"))";var_type=Int;}::env.global else env.global;
	  local=if env.is_global then env.local else {name=f_name;start_str="("^var_req_str^"(*";end_str=")("^str_poi^"))";var_type=Int;}::env.local;
	func_count=env2.func_count;
	func_code=env2.func_code^var_req_str^" function"^(string_of_int (new_env.func_count+1))^"("^str_dec^"){\n"^(if func.return_type.v_type!=Void then(str_of_vartype func.return_type.v_type)^" _return_value;" else "")^block_start^str_head^"_CurSymbolTable->add_function((func_pointer) &function"^(string_of_int (new_env.func_count+1))^",\""^f_name^"\");\n"^stmts^"endline:\n"^block_end^(if func.return_type.v_type!=Void then "return _return_value;" else "") ^"}\n";
	is_global=env.is_global}
	(*(eval_var_req func.return_type)^" "^func.fname^"("^(eval_actual func.formals)^"){\n"^(List.fold_left concat "" (List.map eval_stmt func.body))^"}"*)
and eval_var_req vr=
	(str_of_vartype vr.v_type)^" "^vr.v_name
and eval_var_dec vd env=(**)
	"_CurSymbolTable->add_var("^(eval_vardec_init vd env)^",\""^vd.v_name^"\")",{global=if env.is_global then {name=vd.v_name;start_str="";end_str="";var_type=vd.v_type}::env.global else env.global;local=if env.is_global then env.local else {name=vd.v_name;start_str="";end_str="";var_type=vd.v_type}::env.local;func_count=env.func_count;func_code=env.func_code;is_global=env.is_global}	
and eval expr env = match expr with
	Binop(e1,op,e2) -> "("^(eval e1 env)^(str_of_op op)^(eval e2 env)^")"
	| Call(id,formal)-> let name,str_start,str_end,v_type,is_found =search_var id env in "("^str_start^str_end^"_CurSymbolTable->get_function(\""^id^"\"))"^"("^(eval_formal formal env)^")"
	(*id^"("^(eval_formal formal env)^")"*)(**)
	| Property(id,prop)->let name,str_start,str_end,v_type,is_found =search_var id env in "(("^(str_of_vartype v_type)^")_CurSymbolTable->get_var(\""^id^"\"))"^"->"^prop(**)
	|Property_Call(id,method_name,paras)->let name,str_start,str_end,v_type,is_found =search_var id env in let parameters = eval_formal paras env in ""^(str_of_vartype_call v_type)^")_CurSymbolTable->get_var(\""^id^"\"))"^"->"^method_name^"("^parameters^")"(**)
	| Integer(value)-> string_of_int value
	| String(value)->value
	| Float(value)->string_of_float value
	| Bool(value)->string_of_bool value
	| Id(id)->let name,str_start,str_end,v_type,is_found =search_var id env in (*"*(("^*)(str_of_vartype_call v_type)^")_CurSymbolTable->get_var(\""^id^"\"))"(*id*)(*need to get type!*)
	| Assign(id, e)->let name,str_start,str_end,v_type,is_found =search_var id env in "*(("^(str_of_vartype_2 v_type)^"*)_CurSymbolTable->get_var(\""^id^"\"))="^(eval e env)
	| Noexpr->""
	| Uniop(op,e)->(if op==Uminus then "-" else "!")^(eval e env)
	| Library_Call(lib,method_name,paras)->"library."^method_name^"("^(eval_formal paras env)^")"
and eval_program pr env=
	eval_stmt_list (List.rev pr) env
and eval_actual actual_list env=match actual_list with(*evaluate actuals and give its form*)
	[]->"","","",env(*(int a,int b),(int,int),(_CurSymbolTable->add_var))*)
	|[_]->(eval_var_req (List.hd actual_list)),(str_of_vartype (List.hd actual_list).v_type),"_CurSymbolTable->add_var("^(List.hd actual_list).v_name^",\""^(List.hd actual_list).v_name^"\");\n",{global=env.global; local={name=(List.hd actual_list).v_name;start_str="";end_str="";var_type=(List.hd actual_list).v_type}::env.local;func_count=env.func_count;func_code=env.func_code;is_global=false}
	|h::t -> let x1,y1,z1,w1 =(eval_actual [h] env) in let x,y,z,w =(eval_actual t w1) in  x1^","^x,y1^","^y,z1^z,w
and eval_formal formal_list env=match formal_list with(**)
	[]->""
	|[_]->(eval (List.hd formal_list) env)
	|h::t -> (eval h env)^","^(eval_formal t env)
and eval_vardec_init vd env= match vd with(**)
	{v_type=Cir;init_value=Integer(0)}->"NULL"(*"new Circle"*)
	|{v_type=Lin;init_value=Integer(0)}->"NULL"(*"new Line"*)
	|{v_type=Poi;init_value=Integer(0)}->"NULL"(*"new Point"*)
	|_->(eval vd.init_value env)
and eval_count v = match v with
	|_->v

(*main program*)
	
let _ =
  let head="int main(){\n_sketch=new Sketch;_CurSymbolTable=new Symbol_Table_Node(NULL);\nfopen_s(&_resultfile,\"result.txt\",\"w\");"
  and head1="#include \"stdafx.h\"\n#include <string>\n#include <iostream>\n#include <list>\nusing namespace std;\nlib library;\nSketch* _sketch;\nSymbol_Table_Node* _CurSymbolTable;\nFILE* _resultfile;\n"
  and tail=block_end^"}" 
  and env={global=[];local=[];func_count=0;func_code="";is_global=true} 
  and file_out = open_out "result.cpp" 
  and file = open_in "a.txt" in
  let lexbuf = Lexing.from_channel file in
  let program = Parser.program Scanner.token lexbuf in
  let result,newenv =  eval_program program env in
  ignore ""(*print_endline result*);ignore (output_string file_out (head1^newenv.func_code^head^result^tail));ignore (close_out file_out);;	
(*input_line stdin*)
